# Copyright 2022-2023 The Wordcab Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Live client feature to communicate with a websocket endpoint."""

import asyncio
import json
import logging
import threading
from typing import Optional, Union

import pyaudio
import websockets

from wordcab.login import get_token

logger = logging.getLogger(__name__)


CHUNK = 36000  # Number of audio frames per buffer
FORMAT = pyaudio.paInt16  # Format for audio input (16-bit PCM)
CHANNELS = 1  # Mono
SR = 16000  # Sample rate


class LiveClient:
    """Wordcab API LiveClient used to transcribe audio in real-time."""

    def __init__(
        self, server_url: str, source_lang: str, api_key: Union[str, None] = None
    ) -> None:
        """Initialize LiveClient."""
        self.server_url = server_url
        if "ws://" not in self.server_url:
            raise ValueError(
                "Expected server_url to be a websocket URL."
                "Example: ws://localhost:5001/api/v1/live\n"
                f"But got: {self.server_url}"
            )

        self.source_lang = source_lang

        if api_key is None:
            self.api_key = get_token()
        else:
            self.api_key = api_key

        if self.api_key is None:
            raise ValueError(
                "API Key not found. You must set the WORDCAB_API_KEY environment"
                " variable. Use `wordcab login` to loginto the Wordcab CLI and set the"
                " environment variable."
            )

    async def __aenter__(self) -> "LiveClient":
        """Enter the live client context."""
        self.websocket = await websockets.connect(
            f"{self.server_url}?source_lang={self.source_lang}"
        )
        return self

    async def __aexit__(
        self,
        exception_type: Optional[Union[ValueError, TypeError, AssertionError]],
        exception_value: Optional[Exception],
        traceback: Optional[Exception],
    ) -> None:
        """Exit the live client context."""
        await self.websocket.close()

    async def send_audio(self, audio_data: bytes) -> str:
        """Send audio data to the websocket endpoint."""
        await self.websocket.send(audio_data)

        response = await self.websocket.recv()

        return response


async def cli_live(server_url: str, source_lang: str, api_key: str) -> None:
    """Transcribe audio in real-time."""
    async with LiveClient(server_url, source_lang, api_key) as live_client:
        print("Connected to the live server.")

        p = pyaudio.PyAudio()
        queue = asyncio.Queue()

        # Define function to run in a thread
        def audio_thread():
            stream = p.open(
                format=FORMAT,
                channels=CHANNELS,
                rate=SR,
                input=True,
                frames_per_buffer=CHUNK,
            )
            print("Recording...")

            while not exit_signal.is_set():
                try:
                    audio_data = stream.read(CHUNK, exception_on_overflow=False)
                    loop.call_soon_threadsafe(queue.put_nowait, audio_data)
                except OSError as e:
                    if e.errno == -9981:
                        # Input buffer overflow, let's continue
                        continue
                    else:
                        raise

            stream.stop_stream()
            stream.close()

        exit_signal = threading.Event()
        loop = asyncio.get_event_loop()

        # Start audio thread
        thread = threading.Thread(target=audio_thread, daemon=True)
        thread.start()

        try:
            while True:
                # Get audio data from queue and send
                audio_data = await queue.get()

                json_result = await live_client.send_audio(audio_data)
                if json_result:
                    print(json.loads(json_result)["text"])

        except KeyboardInterrupt:
            print("Recording stopped.")
            exit_signal.set()
            thread.join()

        p.terminate()



# Copyright 2022-2023 The Wordcab Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wordcab API Config variables."""

from pathlib import Path

AVAILABLE_AUDIO_FORMATS = [".flac", ".m4a", ".mp3", ".mpga", ".ogg", ".wav"]
AVAILABLE_GENERIC_FORMATS = [".json", ".txt"]
AVAILABLE_PLAN = ["free", "metered", "paid"]
CONTEXT_ELEMENTS = ["discussion_points", "issue", "keywords", "next_steps", "purpose"]
EXTRACT_AVAILABLE_STATUS = [
    "Deleted",
    "Error",
    "Extracting",
    "ExtractionComplete",
    "ItemQueued",
    "Pending",
    "PreparingExtraction",
]
EXTRACT_PIPELINES = [
    "questions_answers",
    "topic_segments",
    "emotions",
    "speaker_talk_ratios",
]
LIST_JOBS_ORDER_BY = [
    "time_started",
    "time_completed",
    "-time_started",
    "-time_completed",
]
REQUEST_TIMEOUT = 30
SOURCE_LANG = ["de", "en", "es", "fr", "it", "nl", "pt", "sv"]
SOURCE_OBJECT_MAPPING = {
    "generic": "GenericSource",
    "audio": "AudioSource",
    "wordcab_transcript": "WordcabTranscriptSource",
    "assembly_ai": "AssemblyAISource",
    "deepgram": "DeepgramSource",
    "rev_ai": "RevSource",
    "vtt": "VTTSource",
    "youtube": "YoutubeSource",
}
SUMMARIZE_AVAILABLE_STATUS = [
    "Deleted",
    "Error",
    "ItemQueued",
    "Pending",
    "PreparingSummary",
    "PreparingTranscript",
    "Summarizing",
    "SummaryComplete",
    "Transcribing",
    "TranscriptComplete",
]
SUMMARY_LENGTHS_RANGE = [1, 5]
SUMMARY_PIPELINES = ["transcribe", "summarize"]
SUMMARY_TYPES = ["brief", "conversational", "narrative", "no_speaker"]
TARGET_LANG = ["de", "en", "es", "fr", "it", "pt", "sv"]
TRANSCRIBE_LANGUAGE_CODES = [
    "af",
    "am",
    "ar",
    "as",
    "az",
    "ba",
    "be",
    "bg",
    "bn",
    "bo",
    "br",
    "bs",
    "ca",
    "cs",
    "cy",
    "da",
    "de",
    "el",
    "en",
    "es",
    "et",
    "eu",
    "fa",
    "fi",
    "fo",
    "fr",
    "gl",
    "gu",
    "ha",
    "haw",
    "he",
    "hi",
    "hr",
    "ht",
    "hu",
    "hy",
    "id",
    "is",
    "it",
    "ja",
    "jw",
    "ka",
    "kk",
    "km",
    "kn",
    "ko",
    "la",
    "lb",
    "ln",
    "lo",
    "lt",
    "lv",
    "mg",
    "mi",
    "mk",
    "ml",
    "mn",
    "mr",
    "ms",
    "mt",
    "my",
    "ne",
    "nl",
    "nn",
    "no",
    "oc",
    "pa",
    "pl",
    "ps",
    "pt",
    "ro",
    "ru",
    "sa",
    "sd",
    "si",
    "sk",
    "sl",
    "sn",
    "so",
    "sq",
    "sr",
    "su",
    "sv",
    "sw",
    "ta",
    "te",
    "tg",
    "th",
    "tk",
    "tl",
    "tr",
    "tt",
    "uk",
    "ur",
    "uz",
    "vi",
    "yi",
    "yo",
    "zh",
]
TRANSCRIPT_SPEAKER_MAPPING = {
    0: "A",
    1: "B",
    2: "C",
    3: "D",
    4: "E",
    5: "F",
    6: "G",
    7: "H",
    8: "I",
    9: "J",
    10: "K",
    11: "L",
    12: "M",
    13: "N",
    14: "O",
    15: "P",
    16: "Q",
    17: "R",
    18: "S",
    19: "T",
    20: "U",
    21: "V",
    22: "W",
    23: "X",
    24: "Y",
    25: "Z",
}
WORDCAB_TOKEN_FOLDER = Path.home() / ".wordcab" / "token"  # noqa: S105



# Copyright 2022-2023 The Wordcab Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command-line interface."""
import asyncio
from functools import wraps

import click

from .login import cli_login, cli_logout


def coroutine(f) -> asyncio.coroutine:
    """Decorator to run a function as a coroutine."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.group()
@click.version_option()
def main() -> None:
    """Wordcab Python SDK."""
    pass


@click.command()
def login() -> None:
    """Prompt the user for API credentials and store them as git credentials."""
    cli_login()


@click.command()
def logout() -> None:
    """Remove stored git credentials."""
    cli_logout()


main.add_command(login)
main.add_command(logout)

try:
    from .live import cli_live

    @click.command()
    @click.option(
        "--server-url",
        "-s",
        default="ws://localhost:5001/api/v1/live",
        help="Wordcab API Live server URL",
    )
    @click.option(
        "--source-lang",
        "-l",
        default="en",
        help="Source language of the audio",
    )
    @click.option(
        "--api-key",
        "-k",
        default=None,
        help="Wordcab API Key",
    )
    @coroutine
    async def live(server_url: str, source_lang: str, api_key: str) -> None:
        """Transcribe audio in real-time."""
        await cli_live(server_url, source_lang, api_key)

    main.add_command(live)

except ImportError:
    pass


if __name__ == "__main__":
    main(prog_name="wordcab")  # pragma: no cover
